import bpy
from bpy.types import PropertyGroup

from pathlib import Path

from .section_plane import SECTIONBOX_PG_SectionPlane
from .section_box import SECTIONBOX_PG_SectionBox, SECTIONBOX_PG_IncludedObject
from .ui import SECTIONBOX_PG_UI
from .view import SECTIONBOX_PG_View
from ..utils.draw_handles import DrawHandles
from ..utils.sectionbox import get_sectionbox
from ..utils.collection import get_collection
from ..utils.modals import init_modals
from ..utils.data import import_data


def update_active(self, context):
    sb = get_sectionbox(context, check_active=False)
    if not sb:
        return

    # set sb object active
    for obj in context.selected_objects:
        obj.select_set(False)

    sb.object.select_set(True)
    context.view_layer.objects.active = sb.object


class SECTIONBOX_PG_SectionBoxObject(PropertyGroup):
    object: bpy.props.PointerProperty(type=bpy.types.Object)

    def is_valid(self):
        if not self.object:
            return False

        sb = getattr(self.object, "sectionbox", None)
        if not sb:
            return False

        cutter = sb.cutter
        if not cutter:
            return False

        return True

    def delete(self, context):
        if self.object:
            bpy.data.meshes.remove(self.object.data)


class SECTIONBOX_PG_SceneProperties(PropertyGroup):
    ui: bpy.props.PointerProperty(type=SECTIONBOX_PG_UI)
    list: bpy.props.CollectionProperty(type=SECTIONBOX_PG_SectionBoxObject)
    list_index: bpy.props.IntProperty(update=update_active)
    collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    exclude_render_collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    include_render_collection: bpy.props.PointerProperty(type=bpy.types.Collection)

    def create(self, context, objects, collection=None):
        if not self.collection:
            context.scene.sectionbox.collection = get_collection(
                context, f"sb.{context.scene.name}.{context.view_layer.name}"
            )

        # import section node group if not already
        path = Path(__file__).parent.parent / "libs" / "section_node_group.blend"
        import_data(
            str(path),
            "node_groups",
            "section_node_group",
        )

        # create sb object
        sb_id = self.new_id(context)
        sb_obj = self.list.add()
        name = f"sectionbox_{sb_id}"
        sb_obj.name = name
        sb_mesh = bpy.data.meshes.new(name)
        sb_obj.object = bpy.data.objects.new(f"sectionbox_{sb_id}", sb_mesh)

        # create sb data
        sb = sb_obj.object.sectionbox
        sb.id = sb_id
        sb.name = name
        sb.entry_name = f"Section Box{f' {sb.id}' if sb.id != 0 else ''}"
        sb.collection = collection
        sb.object = sb_obj.object
        sb.init(context, objects)

        self.list_index = len(self.list) - 1

        # init handles
        if not bpy.app.driver_namespace.get("sectionbox.draw_handles"):
            bpy.app.driver_namespace["sectionbox.draw_handles"] = DrawHandles(
                bpy.context
            )

        # run modals if not already
        if not context.window_manager.sectionbox.running_modals:
            init_modals(context)

        return sb

    def new_id(self, context):
        # clean invalid sbs from list
        for i, sb_obj in enumerate(context.scene.sectionbox.list):
            if not sb_obj.is_valid():
                # remove sb object
                sb_obj.delete(context)

                # remove sb from list
                context.scene.sectionbox.list.remove(i)

        if not context.scene.sectionbox.list:
            return 0

        id = min([i.object.sectionbox.id for i in context.scene.sectionbox.list])
        if id > 0:
            return id - 1
        else:
            return (
                max([i.object.sectionbox.id for i in context.scene.sectionbox.list]) + 1
            )


class SECTIONBOX_PG_WindowManagerProperties(PropertyGroup):
    running_modals: bpy.props.BoolProperty(default=False)
    edge_select: bpy.props.BoolProperty(default=False)
    hovering_plane: bpy.props.IntProperty(default=-1)
    hovering_edge: bpy.props.IntProperty(default=-1)


classes = [
    SECTIONBOX_PG_WindowManagerProperties,
    SECTIONBOX_PG_SectionBoxObject,
    SECTIONBOX_PG_View,
    SECTIONBOX_PG_SectionPlane,
    SECTIONBOX_PG_IncludedObject,
    SECTIONBOX_PG_SectionBox,
    SECTIONBOX_PG_UI,
    SECTIONBOX_PG_SceneProperties,
]


def register_props():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.types.WindowManager.sectionbox = bpy.props.PointerProperty(
        type=SECTIONBOX_PG_WindowManagerProperties
    )
    bpy.types.Object.sectionbox = bpy.props.PointerProperty(
        type=SECTIONBOX_PG_SectionBox
    )
    bpy.types.Scene.sectionbox = bpy.props.PointerProperty(
        type=SECTIONBOX_PG_SceneProperties
    )


def unregister_props():
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)

    del bpy.types.Scene.sectionbox
    del bpy.types.Object.sectionbox
    del bpy.types.WindowManager.sectionbox
