import bpy
from bpy.types import PropertyGroup

from mathutils import Vector
from functools import cache

from .sectionplane_settings import SECTIONBOX_PG_SectionPlaneSettings
from ..utils.mouse import mouse_hit, location_3d_to_2d
from ..utils.preferences import get_prefs
from ..utils.plane_info import plane_rotation_from_face_normal


class SECTIONBOX_PG_SectionPlane(SECTIONBOX_PG_SectionPlaneSettings, PropertyGroup):
    sb_name: bpy.props.StringProperty()
    id: bpy.props.IntProperty()
    object: bpy.props.PointerProperty(type=bpy.types.Object)
    initial_location: bpy.props.FloatVectorProperty(
        subtype="XYZ", default=(0.0, 0.0, 0.0)
    )
    camera: bpy.props.PointerProperty(type=bpy.types.Object)
    intersecting: bpy.props.BoolProperty()
    pre_settings: bpy.props.PointerProperty(type=SECTIONBOX_PG_SectionPlaneSettings)

    # cross section objects
    cross_section: bpy.props.PointerProperty(type=bpy.types.Object)
    cross_section_gpencil: bpy.props.PointerProperty(type=bpy.types.Object)

    # elevation objects
    elevation: bpy.props.PointerProperty(type=bpy.types.Object)
    elevation_gpencil: bpy.props.PointerProperty(type=bpy.types.Object)

    @cache
    def get_sb(self, context):
        if not self.sb_name:
            return

        sb = context.scene.sectionbox.list.get(self.sb_name)
        return sb.object.sectionbox if sb else None

    def init(self, context):
        sb = self.get_sb(context)
        self.initial_location = sb.object.data.polygons[self.id].center

        # create plane object
        self.object = bpy.data.objects.new(f"{self.sb_name}.{self.id}", None)
        sb.sb_collection.objects.link(self.object)
        self.object.hide_select = True
        self.object.hide_render = True
        self.object.hide_viewport = True
        self.object.parent = sb.object
        self.create_camera(context)
        self.update_transform(context)

    def delete(self, context):
        if self.camera:
            bpy.data.cameras.remove(self.camera.data)
        if self.object:
            bpy.data.objects.remove(self.object)

        self.delete_cross_section(context)

    def create_camera(self, context):
        """Create camera for cross section and elevation."""

        sb = self.get_sb(context)
        cam = bpy.data.cameras.new(f"{self.object.name}_camera")
        cam_obj = bpy.data.objects.new(f"{self.object.name}_camera", cam)
        sb.sb_collection.objects.link(cam_obj)
        cam_obj.parent = self.object
        cam_obj.location = Vector((0, 0, 10))
        cam_obj.data.clip_end = 10.0001
        cam_obj.data.clip_start = 0.1
        cam_obj.data.type = "ORTHO"
        cam_obj.hide_viewport = True
        cam_obj.hide_select = True
        self.camera = cam_obj

    def update_transform(self, context):
        sb = self.get_sb(context)
        face = sb.object.data.polygons[self.id]

        # update plane object location
        self.object.location = face.center

        # update plane object rotation
        self.object.rotation_euler = plane_rotation_from_face_normal[self.id](
            face.normal
        )

        # update camera transform
        mw = sb.object.matrix_world
        self.camera.data.ortho_scale = (
            mw @ face.center - mw @ sb.object.data.vertices[face.vertices[0]].co
        ).length * 2

    def update_intersecting(self, context):
        sb = self.get_sb(context)
        face = sb.object.data.polygons[self.id]
        center = sb.object.matrix_world @ face.center
        aabb = sb.aabb
        cos = [c for c in aabb]
        for i in range(8):
            co = Vector((aabb[i * 3], aabb[i * 3 + 1], aabb[i * 3 + 2]))
            if face.normal.dot(co - center) > 0:
                self.intersecting = True
                return

        self.intersecting = False

    def get_normal(self):
        """Section Box space normal vector of the plane.

        :return: Normal
        :rtype: 'mathutils.Vector'
        """
        return self.object.location.normalized()

    def update_rotation(self):
        for i, face in enumerate(self.sb.object.data.polygons):
            rot = plane_rotation_from_face_normal[i](face.normal)
            self.object.rotation_euler = rot

    def location(self):
        return self.object.parent.matrix_world @ self.object.location

    def location_2d(self, context):
        import bpy_extras

        co = bpy_extras.view3d_utils.location_3d_to_region_2d(
            context.region, context.space_data.region_3d, self.location()
        )

        return co

    def select_edge(self, context, mouse_co):
        prefs = get_prefs()
        sb_object = self.object.parent
        face = sb_object.data.polygons[self.id]
        edges = [
            sb_object.data.edges[sb_object.data.edge_keys.index(ek)]
            for ek in face.edge_keys
        ]
        edge_centers = []
        for edge in edges:
            edge_centers.append(
                (
                    sb_object.data.vertices[edge.vertices[0]].co
                    + sb_object.data.vertices[edge.vertices[1]].co
                )
                / 2
            )

        edge_id = mouse_hit(
            mouse_co,
            [
                location_3d_to_2d(context, sb_object.matrix_world @ ec)
                for ec in edge_centers
            ],
            prefs.select_radius,
        )

        if edge_id is not None:
            return edge_id

    def delete_cross_section(self, context):
        if self.cross_section:
            bpy.data.meshes.remove(self.cross_section.data)
            self.cross_section = None
        if self.cross_section_gpencil:
            for mat in self.cross_section_gpencil.data.materials:
                bpy.data.materials.remove(mat)

            bpy.data.grease_pencils.remove(self.cross_section_gpencil.data)
            self.cross_section_gpencil = None
        self.delete_elevation(context)

    def delete_elevation(self, context):
        if self.elevation:
            bpy.data.meshes.remove(self.elevation.data)
            self.elevation = None
        if self.elevation_gpencil:
            for mat in self.elevation_gpencil.data.materials:
                bpy.data.materials.remove(mat)

            bpy.data.grease_pencils.remove(self.elevation_gpencil.data)
            self.elevation_gpencil = None

        svg_mat = bpy.data.materials.get("SVGMat")
        if svg_mat:
            bpy.data.materials.remove(svg_mat)

    def update_invert(self, invert):
        factor = -1 if invert else 1
        self.camera.location = Vector((0, 0, factor * 10))
        self.camera.scale = Vector((1, 1, factor * 1))
