import bpy

from ..utils.ui import sub_layout, expand_layout, split_layout
from ..utils.sectionbox import get_sectionbox
from .. import bl_info
from .. import addon_id


class SECTIONBOX_PT_MainMenu(bpy.types.Panel):
    bl_label = "Section Box"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Section Box"

    spacing_right = 4.8
    sub1_spacing = 2.5
    split_factor = 0.45

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        # header
        sub = sub_layout(layout, spacing=(0, 0, self.spacing_right), align=True)
        split = sub.split(factor=0.66)
        col = split.column()
        col.enabled = False
        col.label(
            text=f"{bl_info['name']} v{bl_info['version'][0]}.{bl_info['version'][1]}.{bl_info['version'][2]}"
        )
        col = split.column()
        row = col.row(align=True)
        row.alignment = "RIGHT"
        row.operator(
            "preferences.addon_show", text="", icon="PREFERENCES"
        ).module = addon_id
        row.operator(
            "wm.url_open", text="", icon="HELP"
        ).url = f"https://section-box.readthedocs.io/en/latest/"

        # sectionbox list
        row = layout.row()
        row.template_list(
            "SECTIONBOX_UL_List",
            "list",
            scene.sectionbox,
            "list",
            scene.sectionbox,
            "list_index",
        )

        col = row.column(align=True)
        col.operator("sectionbox.list_action", icon="ADD", text="").action = "add"
        row = col.row(align=True)
        row.enabled = (
            scene.sectionbox.list_index < len(context.scene.sectionbox.list) > 0
        )
        row.operator("sectionbox.list_action", icon="REMOVE", text="").action = "remove"
        col.separator()
        col.operator("sectionbox.list_action", icon="TRIA_UP", text="").action = "up"
        col.operator(
            "sectionbox.list_action", icon="TRIA_DOWN", text=""
        ).action = "down"

        col.separator()
        col.menu("SECTIONBOX_MT_ListMenu", text="", icon="DOWNARROW_HLT")

        # no sectionboxes
        if not context.scene.sectionbox.list:
            return

        sb = get_sectionbox(context)
        if not sb:
            return

        # views
        sub0 = sub_layout(layout, spacing=(0, 0, self.spacing_right), align=True)
        row = sub0.row(align=True)
        view_name = "Views" if sb.active_view == -1 else sb.views[sb.active_view].name
        row.menu("SECTIONBOX_MT_ViewMenu", text=view_name)
        row.operator("sectionbox.view", text="", icon="DUPLICATE").action = "override"
        row.operator("sectionbox.view", text="", icon="ADD").action = "add"
        row.operator("sectionbox.view", text="", icon="REMOVE").action = "remove"

        sub0.separator()
        row = expand_layout(
            sub0, "General Settings", 0, scene.sectionbox.ui, "sectionbox_settings"
        )
        if scene.sectionbox.ui.sectionbox_settings:
            sub1 = sub_layout(sub0, spacing=(0, 0, 0), align=True)
            lcol, rcol = split_layout(
                sub1, self.split_factor, self.sub1_spacing, align_right=True
            )
            lcol.label(text="Method")
            rcol.prop(sb, "method", text="")
            if sb.method == "BOOLEAN":
                grid = rcol.grid_flow(
                    row_major=True,
                    columns=2,
                    even_columns=True,
                    even_rows=True,
                    align=True,
                )
                grid.prop(
                    sb,
                    "boolean_use_hole_tolerant",
                    icon="SELECT_DIFFERENCE",
                    text="",
                )
                grid.prop(
                    sb,
                    "boolean_use_self",
                    icon="SELECT_INTERSECT",
                    text="",
                )

            if sb.method == "BOOLEAN":
                lcol, rcol = split_layout(sub1, self.split_factor, self.sub1_spacing)
                lcol.label(text="Material")
                rcol.prop(sb, "material", text="")
            else:
                lcol, rcol = split_layout(sub1, self.split_factor, self.sub1_spacing)
                lcol.label(text="Color")
                rcol.prop(sb, "color", text="")

            lcol, rcol = split_layout(sub1, self.split_factor, self.sub1_spacing)
            lcol.label(text="Invert")
            icon = "CLIPUV_DEHLT" if sb.invert else "REC"
            rcol.prop(sb, "invert", toggle=True, icon=icon, icon_only=True)
            sub1.separator(factor=2)

        # sectionplane settings
        expand_layout(
            sub0, "Plane Settings", 0, scene.sectionbox.ui, "sectionplane_settings"
        )
        if scene.sectionbox.ui.sectionplane_settings:
            if not sb.planes:
                return

            plane = sb.planes[sb.active_plane]

            sub1 = sub_layout(layout, spacing=(0, 0, self.spacing_right), align=True)

            # transform
            lcol, rcol = split_layout(sub1, self.split_factor)
            expand_layout(
                lcol, "Transform", 1, scene.sectionbox.ui, "sectionplane_transform"
            )
            rcol.prop(plane, "transform", text="")
            if scene.sectionbox.ui.sectionplane_transform:
                lcol, rcol = split_layout(
                    sub1, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Edges")
                rcol.prop(plane, "edge_transform", text="")
                sub1.separator(factor=4)

            # cross section
            lcol, rcol = split_layout(sub1, self.split_factor)
            expand_layout(
                lcol, "Section", 1, scene.sectionbox.ui, "sectionplane_cross_section"
            )

            row = rcol.row(align=True)
            text = "Remove" if plane.cross_section else "Create"
            col = row.column()
            op_row = col.row(align=True)
            op_row.enabled = plane.intersecting
            props = op_row.operator("sectionbox.cross_section", text=text)
            props.action = text.lower()
            props.enabled = plane.intersecting
            props = op_row.operator(
                "sectionbox.cross_section", text="", icon="FILE_REFRESH"
            )
            props.action = "update"
            props.enabled = plane.intersecting
            if scene.sectionbox.ui.sectionplane_cross_section:
                # generation settings
                lcol, rcol = split_layout(
                    sub1, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Entire")
                rcol.prop(
                    sb,
                    "cross_section_entire",
                    text="",
                    toggle=True,
                    icon="OBJECT_HIDDEN",
                )
                sub1.separator()

                # display settings
                display_sub = sub1.column(align=True)
                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Distance")
                rcol.prop(sb, "cross_section_dst", text="")
                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Thickness")
                rcol.prop(sb, "cross_section_thickness", text="")
                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Color")
                rcol.prop(sb, "cross_section_color", text="")
                sub1.separator()

                # export settings
                lcol, rcol = split_layout(
                    sub1, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Export as")
                op_row = rcol.row(align=True)
                op_row.prop(scene.sectionbox.ui, "export_format", text="")
                col = op_row.column(align=True)
                col.enabled = bool(plane.cross_section)
                props = col.operator("sectionbox.export", text="", icon="EXPORT")
                props.drawing = "cross_section"
                props.format = scene.sectionbox.ui.export_format
                sub1.separator(factor=4)

            # elevation if cross section existing
            lcol, rcol = split_layout(sub1, self.split_factor)
            expand_layout(
                lcol, "Elevation", 1, scene.sectionbox.ui, "sectionplane_elevation"
            )

            row = rcol.row(align=True)
            col = row.column(align=True)
            col.enabled = bool(plane.cross_section) or not plane.intersecting
            text = "Remove" if plane.elevation else "Create"
            props = col.operator("sectionbox.elevation", text=text)
            props.action = text.lower()
            props.enabled = col.enabled
            row.prop(
                sb,
                "elevation_create_link",
                text="",
                icon="LINKED" if sb.elevation_create_link else "UNLINKED",
            )
            if scene.sectionbox.ui.sectionplane_elevation:
                # generation settings
                lcol, rcol = split_layout(
                    sub1, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Crease")
                rcol.prop(sb, "elevation_crease_angle", text="")
                sub1.separator()

                # display settings
                display_sub = sub1.column(align=True)
                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Distance")
                row = rcol.row(align=True)
                col = row.column(align=True)
                col.enabled = not sb.elevation_dst_link
                col.prop(
                    sb, "elevation_dst" if col.enabled else "cross_section_dst", text=""
                )
                row.prop(
                    sb,
                    "elevation_dst_link",
                    text="",
                    toggle=True,
                    icon="LINKED" if sb.elevation_dst_link else "UNLINKED",
                )

                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Thickness")
                row = rcol.row(align=True)
                col = row.column(align=True)
                col.enabled = not sb.elevation_thickness_link
                col.prop(
                    sb,
                    "elevation_thickness" if col.enabled else "cross_section_thickness",
                    text="",
                )
                row.prop(
                    sb,
                    "elevation_thickness_link",
                    text="",
                    toggle=True,
                    icon="LINKED" if sb.elevation_thickness_link else "UNLINKED",
                )

                lcol, rcol = split_layout(
                    display_sub, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Color")
                row = rcol.row(align=True)
                col = row.column(align=True)
                col.enabled = not sb.elevation_color_link
                col.prop(
                    sb,
                    "elevation_color" if col.enabled else "cross_section_color",
                    text="",
                )
                row.prop(
                    sb,
                    "elevation_color_link",
                    text="",
                    toggle=True,
                    icon="LINKED" if sb.elevation_color_link else "UNLINKED",
                )

                lcol, rcol = split_layout(
                    sub1, self.split_factor, self.sub1_spacing, align_right=True
                )
                lcol.label(text="Export as")
                op_row = rcol.row(align=True)
                op_row.prop(scene.sectionbox.ui, "export_format", text="")
                col = op_row.column(align=True)
                col.enabled = bool(plane.elevation)
                props = col.operator("sectionbox.export", text="", icon="EXPORT")
                props.drawing = "elevation"
                props.format = scene.sectionbox.ui.export_format
