import bpy

from mathutils import Vector

from ..utils.objects import get_evaluated_object, is_geo


class AABB:
    def __init__(self):
        self.min = Vector((1e6, 1e6, 1e6))
        self.max = Vector((-1e6, -1e6, -1e6))

    def from_objects(
        self,
        context,
        objects,
        margin=0,
        empty_size=1,
    ):
        if objects:
            for obj in objects:
                eval_obj = obj.evaluated_get(context.evaluated_depsgraph_get())
                for vert in eval_obj.data.vertices:
                    vert_co = eval_obj.matrix_world @ vert.co
                    for i in range(3):
                        if vert_co[i] < self.min[i]:
                            self.min[i] = vert_co[i]
                        if vert_co[i] > self.max[i]:
                            self.max[i] = vert_co[i]
        else:
            self.min = Vector((-empty_size, -empty_size, -empty_size))
            self.max = Vector((empty_size, empty_size, empty_size))

        self.min -= Vector((margin, margin, margin))
        self.max += Vector((margin, margin, margin))

    def center(self):
        return (self.min + self.max) / 2

    def to_mesh(self, mesh):
        verts = self.get_vertices()
        face_indicies = (
            (3, 2, 1, 0),
            (4, 5, 6, 7),
            (0, 1, 5, 4),
            (2, 3, 7, 6),
            (4, 7, 3, 0),
            (1, 2, 6, 5),
        )
        mesh.from_pydata(verts, [], face_indicies)

    def get_vertices(self):
        return (
            Vector((self.min[0], self.min[1], self.min[2])),
            Vector((self.max[0], self.min[1], self.min[2])),
            Vector((self.max[0], self.max[1], self.min[2])),
            Vector((self.min[0], self.max[1], self.min[2])),
            Vector((self.min[0], self.min[1], self.max[2])),
            Vector((self.max[0], self.min[1], self.max[2])),
            Vector((self.max[0], self.max[1], self.max[2])),
            Vector((self.min[0], self.max[1], self.max[2])),
        )

    def get_intersecting_objects(self, context, location):
        intersecting_objects = []
        for obj in [o for o in context.visible_objects if is_geo(o)]:
            tmp_obj = get_evaluated_object(context, obj)
            if tmp_obj:
                eval_obj = tmp_obj.evaluated_get(context.evaluated_depsgraph_get())
            else:
                eval_obj = obj.evaluated_get(context.evaluated_depsgraph_get())

            for vert in eval_obj.data.vertices:
                vert_co = (eval_obj.matrix_world @ vert.co) - location
                if vert_co.x > self.min.x and vert_co.x < self.max.x:
                    if vert_co.y > self.min.y and vert_co.y < self.max.y:
                        if vert_co.z > self.min.z and vert_co.z < self.max.z:
                            intersecting_objects.append(obj)
                            break

            if tmp_obj:
                bpy.data.meshes.remove(tmp_obj.data)

        return intersecting_objects
