import bpy
import os

from itertools import chain
from pathlib import Path

from .data import import_data
from .preferences import get_prefs
from .properties import prop_filter


plane_order = [
    "bottom",
    "top",
    "front",
    "back",
    "left",
    "right",
]


def clean_objects(context, sb, objects):
    # delete custom prop
    for o in objects:
        del o[sb.name]

    # delete modifier
    for o in objects:
        mod = o.modifiers.get(sb.name)
        if not mod:
            continue

        o.modifiers.remove(mod)

    # delete cross sections if any
    cs_col = bpy.data.collections.get("cross_sections")
    if cs_col:
        for obj in objects:
            [bpy.data.objects.remove(o) for o in cs_col.objects.values() if o.name.endswith(obj.name)]

    # get all materials
    materials = set(ms.material for ms in chain.from_iterable(o.material_slots for o in objects) if ms.material)

    # clean geo nodes
    for o in objects:
        geo_mods = [gm for gm in o.modifiers.values() if gm.type == "NODES"]
        if geo_mods:
            for gm in geo_mods:
                nt: bpy.types.GeometryNodeTree = gm.node_group
                if not nt:
                    continue

                # remove realize instances
                realize_instances = [n for n in nt.nodes if n.name.startswith(f"{sb.name}.realize_instances")]
                if realize_instances:
                    realize_instances = realize_instances[0]
                    prev_socket = realize_instances.inputs[0].links[0].from_socket
                    after_socket = realize_instances.outputs[0].links[0].to_socket
                    nt.links.new(prev_socket, after_socket)
                    nt.nodes.remove(realize_instances)

                # find materials
                for i in [i for i in nt.inputs if i.type == "MATERIAL" and i.default_value]:
                    materials.add(i.default_value)

                for n in [n for n in nt.nodes if n.type == "SET_MATERIAL" and n.inputs[2].default_value]:
                    materials.add(n.inputs[2].default_value)

    for m in materials:
        if m.name.startswith("section_box"):
            del_m = True
            for o in objects:
                if any(p for p in o.keys() if p.startswith("section_box")):
                    del_m = False
                    continue

                context.view_layer.objects.active = o
                idx = [i for i, ms in enumerate(o.material_slots) if ms.material == m]
                if not idx:
                    continue

                o.active_material_index = idx[0]
                bpy.ops.object.material_slot_remove()

            if del_m:
                bpy.data.materials.remove(m)
            continue

        m: bpy.types.Material
        try:
            sb_node = m.node_tree.nodes[sb.name]
        except:
            continue

        output_socket = sb_node.outputs[0].links[0].to_socket
        output_node = sb_node.outputs[0].links[0].to_node
        final_node_socket = sb_node.inputs["Shader"].links[0].from_socket
        final_node = sb_node.inputs["Shader"].links[0].from_node

        # delete texcos
        for input in [i for k, i in sb_node.inputs.items() if k != "Shader" and i.links]:
            texco = input.links[0].from_node
            m.node_tree.nodes.remove(texco)

        m.node_tree.nodes.remove(sb_node)
        m.node_tree.links.new(final_node.outputs[final_node_socket.name], output_node.inputs[output_socket.name])

        # reset blend mode if this was the only section box of the material
        if not [n for n in m.node_tree.nodes if n.name.startswith("section_box")]:
            blm = m.get("blend_method")
            if blm:
                m.blend_method = blm

    # delete tmps
    tmp_col = bpy.data.collections.get("section_box_tmps")
    if tmp_col is not None:
        for o in tmp_col.objects.values():
            orig_o = bpy.data.objects.get(o.name.removesuffix(".section_box_tmp"))
            if orig_o and not any(k.startswith("section_box") for k in o.keys()):
                objects.remove(o)
                bpy.data.objects.remove(o)
                orig_o.hide_set(False)

        if not tmp_col.objects:
            bpy.data.collections.remove(tmp_col)


def delete(context, sb):
    # delete cross sections
    for plane in sb.children:
        delete_cross_sections(sb, plane)

    # delete planes
    for plane in sb.children:
        bpy.data.objects.remove(plane)

    # delete bounding box
    bpy.data.objects.remove(sb)


def delete_cross_sections(sb, plane):
    col = bpy.data.collections.get("cross_sections")
    if col:
        [
            bpy.data.objects.remove(o)
            for o in col.objects
            if o.name.startswith(plane.name.split(".")[0]) and o.name.split(".")[1] == sb.name
        ]
        if not col.objects:
            bpy.data.collections.remove(col)


def convert_actions(old_sbs):
    actions = bpy.data.actions
    affected_actions = []
    for action in actions:
        action: bpy.types.Action
        for fcurve in action.fcurves:
            fcurve: bpy.types.FCurve
            if fcurve.data_path.startswith("sb_"):
                affected_actions.append(action)

    action_groups = {}
    for old_sb in old_sbs:
        old_sb: bpy.types.Object
        if not old_sb.animation_data:
            continue

        if not old_sb.animation_data.action:
            continue

        action = old_sb.animation_data.action
        action_groups[action] = [a for a in affected_actions if old_sb.name in a.name and a is not action]

    for action, plane_actions in action_groups.items():
        for fcurve in action.fcurves:
            fcurve: bpy.types.FCurve
            data_path = fcurve.data_path
            if data_path.startswith("sb_"):
                data_path = data_path.removeprefix("sb_")
                data_path = f"sectionbox.{data_path}"

            fcurve.data_path = data_path

        for plane_action in plane_actions:
            for fcurve in plane_action.fcurves:
                fcurve: bpy.types.FCurve
                data_path = fcurve.data_path
                if data_path.startswith("sb_"):
                    data_path = data_path.removeprefix("sb_")
                if data_path.startswith("plane_"):
                    data_path = data_path.removeprefix("plane_")
                    plane = plane_action.name.split(".")[0]
                    plane_id = plane_order.index(plane)
                    data_path = f"sectionbox.planes[{plane_id}].{data_path}"

                new_fcurve = action.fcurves.new(
                    data_path=data_path,
                )
                for prop in filter(lambda a: prop_filter(fcurve, a), dir(fcurve)):
                    if prop in {"data_path"}:
                        continue
                    try:
                        setattr(new_fcurve, prop, getattr(fcurve, prop))
                    except:
                        continue

                for kf in fcurve.keyframe_points:
                    kf: bpy.types.Keyframe
                    new_kf = new_fcurve.keyframe_points.insert(
                        kf.co[0],
                        kf.co[1],
                    )
                    for prop in filter(lambda a: prop_filter(kf, a), dir(kf)):
                        setattr(new_kf, prop, getattr(kf, prop))

            bpy.data.actions.remove(plane_action)


def convert_sbs(context, old_sbs, obj_list):
    prefs = get_prefs()
    scene = context.scene

    # get areas
    view_3d = None
    outliner = None
    for area in context.screen.areas:
        if area.type == "VIEW_3D":
            view_3d = area
        elif area.type == "OUTLINER":
            outliner = area

    # disable margin
    prev_margin = prefs.margin
    prefs.margin = 0.0

    # create new section boxes and delete old ones
    for i, old_sb in enumerate(old_sbs):
        try:
            collection = bpy.data.collections.get(old_sb["collection"])
        except:
            collection = None

        objects = obj_list[i]
        with context.temp_override(area=outliner if collection else view_3d):
            sb = scene.sectionbox.create(context, objects, collection)

            sb.object.location = sb.cutter.location = old_sb.location
            sb.object.rotation_euler = sb.cutter.rotation_euler = old_sb.rotation_euler
            sb.object.scale = sb.cutter.scale = old_sb.scale

            sb.material = old_sb.active_material.name if old_sb.active_material else "From Mesh"
            sb.color = old_sb.sb_color[:3]
            sb.method = old_sb.sb_method
            sb.invert = old_sb.sb_invert
            sb.enable = old_sb.sb_enabled
            sb.hide = old_sb.sb_hide
            sb.hide_render = old_sb.sb_hide_render
            old_sb_item = [i for i in context.scene.sb_list.values() if i.object == old_sb]
            if old_sb_item:
                sb.entry_name = old_sb_item[0].name

            for old_plane, plane in zip(old_sb["planes"], sb.planes):
                plane.transform = old_plane.sb_plane_transform

            try:
                views = old_sb.sb_views
            except:
                pass
            else:
                for old_view in views:
                    view = sb.views.add()
                    view.name = old_view.name
                    sb_settings = view.sectionbox_settings
                    sb_settings.method = "BOOLEAN"
                    sb_settings.invert = old_view.invert
                    sb_settings.hide = False
                    sb_settings.hide_render = False
                    sb_settings.enable = True
                    for i in range(6):
                        plane_settings = view.plane_settings.add()
                        plane_settings.transform = old_view.plane_transforms[i]

            # apply old action
            if old_sb.animation_data and old_sb.animation_data.action:
                old_action = old_sb.animation_data.action
                if old_action:
                    if not sb.object.animation_data:
                        sb.object.animation_data_create()

                    sb.object.animation_data.action = old_action

    # convert old actions
    convert_actions(old_sbs)

    # restore margin
    prefs.margin = prev_margin

    # set active section box
    old_idx = context.scene.get("sb_list_index")
    context.scene.sectionbox.list_index = old_idx


def convert_old(context):
    """Clean up old sb data and convert old sbs if any"""

    scene = context.scene

    # check if old sbs exist
    sb_list = scene.get("sb_list")
    if not sb_list:
        return

    # remove old driver namespace
    dh = bpy.app.driver_namespace.get("draw_handles")
    if dh:
        dh.remove_handler()
        del bpy.app.driver_namespace["draw_handles"]

    # get old section boxes and clean objects
    old_sb_collection = bpy.data.collections.get("section_boxes." + scene.name)
    old_sbs = [o for o in old_sb_collection.objects if o.type == "MESH" and o.name.startswith("section_box")]
    obj_list = []
    for old_sb in old_sbs:
        objects = [o for o in scene.objects if o.get(old_sb.name) is not None and o is not old_sb]
        obj_list.append(objects)
        clean_objects(context, old_sb, objects)

    # convert old sbs if any
    if bpy.context.preferences.addons.get("Section Box"):
        convert_sbs(context, old_sbs, obj_list)

    # delete old section boxes if any
    for old_sb in old_sbs:
        delete(context, old_sb)

    # remove old data if any
    old_sb_node_group = bpy.data.node_groups.get("Section Box")
    if old_sb_node_group:
        bpy.data.node_groups.remove(old_sb_node_group)

    try:
        del scene["sb_list"]
        del scene["sb_list_index"]
        del scene["sb_ui"]
        del scene["shift"]
    except Exception as e:
        print(e)

    # force section box transform update
    for sb_obj in scene.sectionbox.list:
        sb = sb_obj.object.sectionbox
        for plane in sb.planes:
            sb.active_plane = plane.id
            plane.transform = plane.transform

    if old_sb_collection:
        bpy.data.collections.remove(old_sb_collection)


def convert_keymaps():
    context = bpy.context

    def update_kmi(km, old_kmi, new_kmi_idname):
        props = ("map_type", "type", "alt", "shift", "ctrl", "oskey", "any")
        new_kmi = context.window_manager.keyconfigs.user.keymaps["Object Mode"].keymap_items.get(new_kmi_idname)
        if not new_kmi:
            return

        for prop in props:
            setattr(new_kmi, prop, getattr(old_kmi, prop))

    # update keymaps
    kc = context.window_manager.keyconfigs
    for km in kc.user.keymaps:
        km: bpy.types.KeyMap
        for kmi in km.keymap_items:
            kmi: bpy.types.KeyMapItem
            if kmi.idname == "sb.select_section_box":
                update_kmi(km, kmi, "sectionbox.select")
            elif kmi.idname == "sb.reset":
                update_kmi(km, kmi, "sectionbox.reset")
            elif kmi.idname == "sb.delete_section_box_hotkey":
                update_kmi(km, kmi, "sectionbox.delete")


def convert_prefs():
    prefs = get_prefs()
    try:
        old_prefs = bpy.context.preferences.addons["Section Box"].preferences
    except:
        return

    # convert default material
    path = Path(bpy.utils.user_resource("SCRIPTS", path="presets"))
    file_path = path.joinpath("section_box", "default_mat.blend")
    if file_path.exists():
        old_def_mat = old_prefs.default_material
        if old_def_mat != "From Mesh" and old_def_mat not in bpy.data.materials.keys():
            materials = import_data(str(file_path), "materials", None)
            prefs.def_material = materials[0].name
        else:
            prefs.def_material = old_def_mat

        prefs.export_def_material()
        file_path.unlink()
        file_path.parent.rmdir()

    # convert other preferences
    prefs.show_menu = old_prefs.show_menu
    prefs.lazy_load = old_prefs.lazy_load
    prefs.def_settings.color = old_prefs.default_color[:3]
    prefs.empty_size = old_prefs.default_empty_sb_size
    prefs.merge_panels = old_prefs.merge_panels


def convert_info():
    context = bpy.context
    window = None
    area = None
    for win in context.window_manager.windows:
        for a in win.screen.areas:
            if a.type == "PREFERENCES":
                area = a
                window = win
                break

        if window:
            break

    if not (window or area):
        return

    with context.temp_override(window=window, area=area):
        bpy.ops.sectionbox.convert_info("INVOKE_DEFAULT")
